/*
*  MIMEmanDec.cpp
*  Copyright 1999-2003 by Benchmark Software
*  All Rights Reserved
*
*  This source file is released for peer review purposes only. It is not public domain. 
*  Redistribution either in part or whole is forbidden. The sole source for QuickSilver 
*  distribution currently is http://quicksilver.skuz.net 
*
*  Richard
*  --
*  R.Christman
*  Benchmark Software
*  quicksilver@skuz.net
*  http://quicksilver.skuz.net
*
*/
//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop

#include "MIMEmanDec.h"
#include "FileMan.h"
#include "global.h"

TMIMEdecoder *MIMEdecoder;
//---------------------------------------------------------------------------
bool __fastcall DecodeMIMEMessage(TStrings *Message)
{
   MIMEdecoder = new TMIMEdecoder(Message);
   bool ok = MIMEdecoder->DecodeFinish()==mrOk;
   delete MIMEdecoder;
   return ok;
}
//---------------------------------------------------------------------------
bool __fastcall DecodeMIMEAttachments(TForm *Owner, TStrings *Message, AnsiString &AttachDir)
{
   MIMEdlg = new TMIMEdlg(Owner,"Decoding Attachment");
   MIMEdecoder = new TMIMEdecoder(MIMEdlg,Message,AttachDir);
   bool ok = MIMEdlg->DecodeAttachments(MIMEdecoder)==mrOk;
   delete MIMEdecoder;
   delete MIMEdlg;
   return ok;
}
//---------------------------------------------------------------------------
bool __fastcall DeleteMIMEImages(TStrings *Message)
{
   MIMEdecoder = new TMIMEdecoder(Message);
   bool ok = MIMEdecoder->DeleteType("image")==mrOk;
   delete MIMEdecoder;
   return ok;
}
//---------------------------------------------------------------------------
bool __fastcall DeleteMIMEAudios(TStrings *Message)
{
   MIMEdecoder = new TMIMEdecoder(Message);
   bool ok = MIMEdecoder->DeleteType("audio")==mrOk;
   delete MIMEdecoder;
   return ok;
}
//---------------------------------------------------------------------------
bool __fastcall DeleteMIMEApplications(TStrings *Message)
{
   MIMEdecoder = new TMIMEdecoder(Message);
   bool ok = MIMEdecoder->DeleteType("application")==mrOk;
   delete MIMEdecoder;
   return ok;
}
//---------------------------------------------------------------------------
// TMIMEdecoder
//---------------------------------------------------------------------------
__fastcall TMIMEdecoder::TMIMEdecoder(TStrings *Message)
{
   mime = Message;
   contenthead = new TStringList();
   head = new TStringList();
   output = new TStringList();
   buffer = new char[1024];
   status = statMimeOk;
   ok = true;
}
//---------------------------------------------------------------------------
__fastcall TMIMEdecoder::TMIMEdecoder(TMIMEdlg *MimeDlg, TStrings *Message,
   AnsiString &AttachDir)
{
   status = statMimeOk; ok = true;
   mime = Message;
   mimedlg = MimeDlg;
   attachdir = AddBS(AttachDir);
   contenthead = new TStringList();
   head = new TStringList();
   output = new TStringList();
   buffer = new char[1024];
}
//---------------------------------------------------------------------------
__fastcall TMIMEdecoder::~TMIMEdecoder()
{
   delete buffer;
   for (int i=0; i<output->Count; i++)
      delete output->Objects[i];
   delete output;
   delete head;
   delete contenthead;
}
//---------------------------------------------------------------------------
bool __fastcall TMIMEdecoder::DecodeFinish()
{
   return DecodeFirstPart();
}
//---------------------------------------------------------------------------
void __fastcall TMIMEdecoder::ShowProgress(double position)
{
   if (mimedlg) mimedlg->ProgressBar->Position = position*100;
}
//---------------------------------------------------------------------------
int __fastcall TMIMEdecoder::GetPartSize(int start, AnsiString &boundary)
{
   if (!mimedlg) return 0;
   double total=0; for (int i=start; i<mime->Count; i++)
   {
      if (mime->Strings[i].LowerCase()=="--"+boundary) break;
      total += mime->Strings[i].Length()+2;
   }
   return total;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TMIMEdecoder::GetContent(AnsiString &Field)
{
   AnsiString item;
   int i=-1; while(!(item=GetToken(++i,content,";").Trim()).IsEmpty())
   {
      if (item.Trim().LowerCase().Pos(Field)==1)
         return dropquotes(item.SubString(Field.Length()+1,1024).Trim());
   }
   return "";
}
//---------------------------------------------------------------------------
bool __fastcall IsBoundary(AnsiString &line, AnsiString &boundary)
{
   return line.LowerCase()=="--"+boundary.LowerCase()+"--" ||
          line.LowerCase()=="--"+boundary.LowerCase() ||
         (line.IsEmpty() && boundary.IsEmpty());
}
//---------------------------------------------------------------------------
int __fastcall TMIMEdecoder::GetAttachmentSize(int start, AnsiString &boundary)
{
   AnsiString line;
   if (!mimedlg) return 0;
   double total=0; for (int i=start+1; i<mime->Count; i++)
   {
      line = mime->Strings[i];
      if (line.IsEmpty() || IsBoundary(line,boundary)) break;
      total += line.Length()+2;
   }
   if (total==0) total = 1;
   return total;
}
//---------------------------------------------------------------------------
void __fastcall TMIMEdecoder::SetAttachedHeader(AnsiString &attached, TStrings *headers)
{
   if (headers->Text.Pos("\r\nAttachment: ")) headers->Add(" "+attached+";");
   else headers->Add("Attachment: "+attached+";");
}
//---------------------------------------------------------------------------
AnsiString __fastcall TMIMEdecoder::GetValidUniqueFilename(AnsiString &filename)
{
   AnsiString ext,xstr,path=AddBS(attachdir);
   if (filename.IsEmpty()) filename = "noname.???";
   // correct name only, valid path provided by QS
   for (int i=1; i<=filename.Length(); i++)
      if (s("\/:*?\"<>|").Pos(filename[i])>0) filename[i] = '_';
   // parse ext and name without ext
   ext = ExtractFileExt(filename);
   filename = filename.SubString(1,filename.Length()-ext.Length());
   int x=0; while(FileExists(path+filename+xstr+ext)) xstr = s(++x);
   return path+filename+xstr+ext;
}
//---------------------------------------------------------------------------
bool __fastcall TMIMEdecoder::ConstructMessage()
{
   mime->Text = head->Text+"\r\n";
   bool hastext=false,hashtml=false;
   for (int i=0; i<output->Count; i++)
   {
      if (output->Strings[i].Pos("text/") && !output->Strings[i].Pos("text/html"))
      {
         mime->AddStrings((TStringList*)output->Objects[i]);
         hastext = true;
      }
   }
   if (!hastext)
   {
      for (int i=0; i<output->Count; i++)
      {
         if (output->Strings[i].Pos("text/html")>0)
         {
            TStrings *lines=(TStringList*)output->Objects[i];
            StripHTML(lines);
            mime->AddStrings(lines);
            hashtml = true;
         }
      }
   }
   return hashtml;
}
//---------------------------------------------------------------------------
bool __fastcall TMIMEdecoder::SeekBoundary(AnsiString &boundary)
{
   while (mime->Count)
   {
      AnsiString line=mime->Strings[0]; mime->Delete(0);
      if (IsBoundary(line,boundary))
         return true;
   }
   return false;
}
//---------------------------------------------------------------------------
//bool __fastcall IsTypeText(AnsiString contenttype)
//{
//   return contenttype=="text/plain" || contenttype=="text/html";
//}
//---------------------------------------------------------------------------
bool __fastcall IsTypeText(AnsiString contenttype)
{
   return contenttype.Pos("text/")==1;
}
//---------------------------------------------------------------------------
bool __fastcall TMIMEdecoder::DecodeFirstPart()
{
   ok = true; status = statMimeOk;
   mime->BeginUpdate();
   version=GetMIMEVersion();
   content=GetPrimeMIMEContent();
   AnsiString line,encoding,pgptail,type,boundary=GetContent("boundary=");
   // save headers
   while (mime->Count)
   {
      line = mime->Strings[0];
      if (line.Trim().IsEmpty() && mime->Count>1 &&
         mime->Strings[1].Pos("*** PGP Signature Status:")==0)
            break;
      mime->Delete(0);
      head->Add(line);
   }
   TrimNullsAfter(mime);
   if (mime->Count && (line=mime->Strings[mime->Count-1]).Pos("*** END PGP ")==1)
      pgptail = line;
   TrimNullsAfter(mime);

   // handle multipart
   if (!boundary.IsEmpty())
   {
      while (DecodeNextPart(boundary))
         continue;

      ConstructMessage();
      if (!pgptail.IsEmpty() && mime->Strings[mime->Count-1]!=pgptail)
         mime->Add(pgptail);
      mime->EndUpdate();
      return ok;
   }
   type=GetContent("content-type: ");
   encoding = GetContent("content-transfer-encoding: ");
   double bytesread=0,bytestotal=GetPartSize(0,"");
   TMemoryStream *stream = new TMemoryStream();
   while (mime->Count)
   {
      AppProcessMessages();
      line = mime->Strings[0];
      bytesread += line.Length()+2;
      if (bytestotal) ShowProgress(bytesread/bytestotal);

      if (encoding=="quoted-printable") DecodeQuotedPrintableLine(line,stream); else
      if (encoding=="x-uuencode") DecodeUUencodeLine(line,stream); else
      if (encoding=="base64") DecodeBase64Line(line,stream); else
         DecodeUncodedLine(line,stream);
      mime->Delete(0);
   }

   stream->Position = 0; output->LoadFromStream(stream);
   if (type.Pos("text/html")) StripHTML(output);
   mime->Text = head->Text+"\r\n";
   mime->AddStrings(output);

   if (!pgptail.IsEmpty() && mime->Strings[mime->Count-1]!=pgptail)
      mime->Add(pgptail);

   AppProcessMessages();
   if (mimedlg)
   {
      Snooze(300);
      ShowProgress(0);
      Snooze(300);
   }
   delete stream;
   mime->EndUpdate();
   return ok;
}
//---------------------------------------------------------------------------
bool __fastcall TMIMEdecoder::DecodeNextPart(AnsiString &Boundary)
{
   TrimNulls(mime);
   AnsiString line,encoding,disposition,boundary,filename,type;
   if (mime->Count==0) return false;
   if (mime->Count==1 && IsBoundary(mime->Strings[0],Boundary))
   {
      mime->Delete(0);
      return false;
   }
   ok = true; status = statMimeOk;
   bool attached;

   content=GetMIMEContent();
   disposition = GetContent("content-disposition: "); attached = disposition=="attachment";
   if (attached) filename = GetValidUniqueFilename(GetContent("filename="));
   boundary = GetContent("boundary="),type=GetContent("content-type: ");
   encoding = GetContent("content-transfer-encoding: ");
   SeekBoundary(Boundary);
   TrimNulls(mime);

   if (!boundary.IsEmpty())
   {
      while (DecodeNextPart(boundary)) continue;
      return false;
   }

   TStream *stream;
   double bytesread=0,bytestotal=GetPartSize(0,Boundary);
   if (attached) stream = new TFileStream(filename,fmCreate);
   else stream = new TMemoryStream();

   while(mime->Count)
   {
      AppProcessMessages();
      line = mime->Strings[0];
      bytesread += line.Length()+2;
      if (bytestotal) ShowProgress(bytesread/bytestotal);
      if (IsBoundary(line,Boundary)) break;
      mime->Delete(0);

      if (encoding=="quoted-printable") DecodeQuotedPrintableLine(line,stream); else
      if (encoding=="x-uuencode") DecodeUUencodeLine(line,stream); else
      if (encoding=="base64") DecodeBase64Line(line,stream); else
         DecodeUncodedLine(line,stream);
   }
   if (attached) SetAttachedHeader(filename,head);
   else
   {
      if (content.Pos("text/")>0)
      {
         TStringList *out = new TStringList();
         stream->Position = 0; out->LoadFromStream(stream);
         output->AddObject(content,out);
      }
   }

   AppProcessMessages();
   if (mimedlg)
   {
      ShowProgress(100);
      Snooze(300);
      ShowProgress(0);
      Snooze(300);
   }
   TrimNulls(mime);
   delete stream;
   return true;
}
//---------------------------------------------------------------------------
/*bool __fastcall TMIMEdecoder::DecodeFirstPart()
{
   ok = true; status = statMimeOk;
   mime->BeginUpdate();

   version=GetMIMEVersion();
   content=GetMIMEContent();
   AnsiString line,encoding,type,boundary=GetContent("boundary=");

   // save headers
   while (mime->Count)
   {
      line = mime->Strings[0];
      if (line.Trim().IsEmpty())
         break;
      mime->Delete(0);
      head->Add(line);
   }
   // handle multipart
   if (!boundary.IsEmpty())
   {
      while (DecodeNextPart(boundary))
         continue;

      ConstructMessage();
      mime->EndUpdate();
      return ok;
   }
   type=GetContent("content-type: ");
   encoding = GetContent("content-transfer-encoding: ");
   double bytesread=0,bytestotal=GetPartSize(0,"");
   TMemoryStream *stream = new TMemoryStream();

   while (mime->Count)
   {
      AppProcessMessages();
      line = mime->Strings[0];
      bytesread += line.Length()+2;
      if (bytestotal) ShowProgress(bytesread/bytestotal);

      if (encoding=="quoted-printable") DecodeQuotedPrintableLine(line,stream); else
      if (encoding=="x-uuencode") DecodeUUencodeLine(line,stream); else
      if (encoding=="base64") DecodeBase64Line(line,stream); else
         DecodeUncodedLine(line,stream);
      mime->Delete(0);
   }

   stream->Position = 0; output->LoadFromStream(stream);
   mime->Text = head->Text+"\r\n";
   mime->AddStrings(output);
   AppProcessMessages();
   if (mimedlg)
   {
      Snooze(300);
      ShowProgress(0);
      Snooze(300);
   }
   delete stream;
   mime->EndUpdate();
   return ok;
}
//---------------------------------------------------------------------------
bool __fastcall TMIMEdecoder::DecodeNextPart(AnsiString &Boundary)
{
   TrimNulls(mime);
   AnsiString line,encoding,disposition,boundary,filename,type;

   if (mime->Count==0) return false;
   if (mime->Count==1 && IsBoundary(mime->Strings[0],Boundary))
   {
      mime->Delete(0);
      return false;
   }
   ok = true; status = statMimeOk;
   bool attached;

   content=GetMIMEContent();
   disposition = GetContent("content-disposition: "); attached = disposition=="attachment";
   if (attached) filename = GetValidUniqueFilename(GetContent("filename="));
   boundary = GetContent("boundary="),type=GetContent("content-type: ");
   encoding = GetContent("content-transfer-encoding: ");
   SeekBoundary(Boundary);
   TrimNulls(mime);

   if (!boundary.IsEmpty())
   {
      while (DecodeNextPart(boundary)) continue;
      return false;
   }

   TStream *stream;
   double bytesread=0,bytestotal=GetPartSize(0,Boundary);
   if (attached) stream = new TFileStream(filename,fmCreate);
   else stream = new TMemoryStream();

   while(mime->Count)
   {
      AppProcessMessages();
      line = mime->Strings[0];
      bytesread += line.Length()+2;
      if (bytestotal) ShowProgress(bytesread/bytestotal);
      if (IsBoundary(line,Boundary)) break;
      mime->Delete(0);

      if (encoding=="quoted-printable") DecodeQuotedPrintableLine(line,stream); else
      if (encoding=="x-uuencode") DecodeUUencodeLine(line,stream); else
      if (encoding=="base64") DecodeBase64Line(line,stream); else
         DecodeUncodedLine(line,stream);
   }
   if (attached) SetAttachedHeader(filename,head);
   else
   {
      if (content.Pos("text/")>0)
      {
         TStringList *out = new TStringList();
         stream->Position = 0; out->LoadFromStream(stream);
         output->AddObject(content,out);
      }
   }

   AppProcessMessages();
   if (mimedlg)
   {
      ShowProgress(100);
      Snooze(300);
      ShowProgress(0);
      Snooze(300);
   }
   TrimNulls(mime);
   delete stream;
   return true;
}*/
//---------------------------------------------------------------------------
const AnsiString encodes=",x-uuencode,base64,7bit,8bit,";
bool __fastcall TMIMEdecoder::DecodeAttachments()
{
   head->Clear();
   AnsiString line,encoding,disposition,boundary,filename,type;
   while (mime->Count)
   {
      line = mime->Strings[0];
      if (line.IsEmpty()) break;
      mime->Delete(0);
      head->Add(line);
   }

   int i=-1; while (++i<mime->Count)
   {  // find first content-disposition line in message
      if (mime->Strings[i].Pos("Content-Disposition: attachment")>0)
      {
         content = "";
         output->Clear();
         contenthead->Clear();
         // backup to beginning of content (with or without boundary)
         while (--i>0)
         {
            line = mime->Strings[i].Trim();
            if (line.IsEmpty() || line.Pos("--")==1) break;
         }

         boundary = mime->Strings[i].Trim();
         if (boundary.Pos("--")==1) boundary = boundary.SubString(3,256);
         else if (boundary.IsEmpty()) i++;

         while (i<mime->Count)
         {
            AppProcessMessages();
            line = mime->Strings[i].Trim();
            if (i<mime->Count-1 && line.IsEmpty() && !mime->Strings[i+1].IsEmpty())
               break;

            AddSemiColon(content); content += " "+line.Trim();
            contenthead->Add(line);
            mime->Delete(i);
         }

         encoding = GetContent("content-transfer-encoding: ");
         filename = GetValidUniqueFilename(GetContent("filename="));
         if (mimedlg) mimedlg->Filename->Caption = "File: "+filename;

         TStream *stream = new TFileStream(filename,fmCreate);
         double bytesread=0,bytestotal=GetAttachmentSize(i++,boundary);
         while (i<mime->Count)
         {
            if (cancel) break;
            AppProcessMessages();
            line = mime->Strings[i];
            bytesread += line.Length()+2;
            ShowProgress(bytesread/bytestotal);

            if (line.Trim().IsEmpty() || IsBoundary(line,boundary)) break;

            if (encoding=="x-uuencode") DecodeUUencodeLine(line,stream); else
            if (encoding=="base64") DecodeBase64Line(line,stream); else
               DecodeUncodedLine(line,stream);
            output->Add(line);
            mime->Delete(i);
         }

         SetAttachedHeader(filename,head);
         if (i>0 && mime->Strings[i-1].IsEmpty())
             mime->Delete(i-1);
         delete stream;

         if (cancel)
         {
            if (FileExists(filename)) DeleteFile(filename.c_str());
            break;
         }
         if (mimedlg)
         {
            ShowProgress(100);
            Snooze(300);
            ShowProgress(0);
            Snooze(300);
         }
         if (!encoding.IsEmpty() && encodes.Pos(","+encoding+",")==0)
         {
            BingBox(AppMainForm,
               s("I have encountered a problem with this attachment:\n")+
               s( filename+"\n\n")+
               s("It uses an encoding scheme I do not support. Consequently,\n")+
               s("I have saved the attachment undecoded with it's content\n")+
               s("headers intact."),
               s("Unknown Encoding"),MB_OK
            );
         }
      }
      if (cancel) break;
      if (encoding=="x-uuencode" && mime->Strings[i-1]=="end")
      {
         mime->Delete(i-1);
         mime->Delete(i-1);
      }
   }

   for (int i=0; i<head->Count; i++)
      mime->Insert(i,head->Strings[i]);
   head->Clear();

   TrimNullsAfter(mime);
   if (mimedlg) mimedlg->ModalResult = cancel?mrCancel:mrOk;
   return true;
}
//---------------------------------------------------------------------------
/*bool __fastcall TMIMEdecoder::DeleteImages()
{
   AnsiString line,boundary;
   int i=-1; while (++i<mime->Count)
   {  // find first content-disposition line in message
      if (mime->Strings[i].Pos("Content-Type: image/")>0)
      {
         // backup to beginning of content (with or without boundary)
         while (i>0 && !(line=mime->Strings[i]).IsEmpty() && line.Pos("--")!=1) i--;
         boundary = mime->Strings[i];
         if (boundary.IsEmpty())
            i++;

         while (i<mime->Count)
         {
            if (cancel) break;
            line = mime->Strings[i];
            mime->Delete(i);
            if (line.Trim().IsEmpty())
               break;
         }
         while (i<mime->Count)
         {
            if (cancel) break;
            line = mime->Strings[i];
            if (line.Trim().IsEmpty() || IsBoundary(line,boundary))
               break;
            mime->Delete(i);
         }

         mime->Delete(i);
         //mime->Delete(i); // delete last boundary
      }
      if (cancel) break;
   }
   if (mimedlg) mimedlg->ModalResult = cancel?mrCancel:mrOk;
   return true;
}
//---------------------------------------------------------------------------
bool __fastcall TMIMEdecoder::DeleteAudios()
{
   AnsiString line,boundary;
   int i=-1; while (++i<mime->Count)
   {  // find first content-disposition line in message
      if (mime->Strings[i].Pos("Content-Type: audio/")>0)
      {
         // backup to beginning of content (with or without boundary)
         while (--i>0)
         {
            line = mime->Strings[i].Trim();
            if (line.IsEmpty() || line.Pos("--")==1) break;
         }

         boundary = mime->Strings[i].Trim();
         if (boundary.Pos("--")==1) boundary = boundary.SubString(3,256);
         else if (boundary.IsEmpty()) i++;

         while (i<mime->Count)
         {
            if (cancel) break;
            line = mime->Strings[i];
            mime->Delete(i);
            if (line.Trim().IsEmpty())
               break;
         }
//view(mime);
         while (i<mime->Count)
         {
            if (cancel) break;
            line = mime->Strings[i];
            if (line.Trim().IsEmpty() || IsBoundary(line,boundary))
               break;
            mime->Delete(i);
         }

         mime->Delete(i);
         //mime->Delete(i); // delete last boundary
      }
      if (cancel) break;
   }
   if (mimedlg) mimedlg->ModalResult = cancel?mrCancel:mrOk;
   return true;
}*/
//---------------------------------------------------------------------------
bool __fastcall TMIMEdecoder::DeleteType(AnsiString &contenttype)
{
   AnsiString line,boundary;
   int i=-1; while (++i<mime->Count)
   {  // find first content-disposition line in message
      if (mime->Strings[i].Pos("Content-Type: "+contenttype+"/")>0)
      {
         // backup to beginning of content (with or without boundary)
         while (--i>0)
         {
            line = mime->Strings[i].Trim();
            if (line.IsEmpty() || line.Pos("--")==1) break;
         }

         boundary = mime->Strings[i].Trim();
         if (boundary.Pos("--")==1) boundary = boundary.SubString(3,256);
         else if (boundary.IsEmpty()) i++;

         while (i<mime->Count)
         {
            if (cancel) break;
            line = mime->Strings[i];
            mime->Delete(i);
            if (line.Trim().IsEmpty())
               break;
         }
         while (i<mime->Count)
         {
            if (cancel) break;
            line = mime->Strings[i];
            if (line.Trim().IsEmpty() || IsBoundary(line,boundary))
               break;
            mime->Delete(i);
         }

         //mime->Delete(i);
         //if (mime->Count>i) mime->Delete(i); // delete last boundary
      }
      if (cancel) break;
   }
   if (mimedlg) mimedlg->ModalResult = cancel?mrCancel:mrOk;
   return true;
}
//---------------------------------------------------------------------------
bool __fastcall TMIMEdecoder::DecodeQuotedPrintableLine(AnsiString &line, TStream *stream)
{
   if (mimedlg && mimedlg->Caption.Pos("Quoted")==0) mimedlg->Caption = "Decoding quoted-printable...";
   strcpy(buffer,s(line+"\r\n").c_str());
   int size=DecodeQuotedPrintableLine(buffer);
   stream->WriteBuffer(buffer,size);
   return status==statMimeOk;
}
//---------------------------------------------------------------------------
#define SKIP '\202'
const char hexmap[] = {
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
      0 ,    1 ,    2 ,    3 ,    4 ,    5 ,    6 ,    7 ,
      8 ,    9 ,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,    10,    11,    12,    13,    14,    15,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,
     SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP,   SKIP
};
int  __fastcall TMIMEdecoder::DecodeQuotedPrintableLine(char *line)
{
  ok = true; status = statMimeOk;
  char *s = line;
  char *finalresult = (char*)calloc(strlen(line) + sizeof(char), sizeof(char));
  char *result = finalresult;
  while (*s != '\0') //loop through the entire string...
  {
    if (cancel) return (ok=false);

    if (*s == '=') //woops, needs to be decoded...
    {
      for (int i = 0; i < 3; i++) //is s more than 3 chars long...
      {
        if (s[i] == '\0')
        {
          //error in the decoding...
          //MsgDlg(mimedlg,
          //   s("Encountered corruption while decoding 'quoted-printable'\n")+
          //   s("MIME part. Skipping..."),"MIME Error",MB_OK);
          //status = statMimeDecodeError;
          //return ok;
        }
      }
      char mid[3];
      s++; //move past the "="
      //let's put the hex part into mid...
      ok = true;
      for (int i = 0; i < 2; i++)
      {
        if (hexmap[s[i]] == SKIP)
        {
          //we have an error, or a linebreak, in the encoding...
          ok = false;
          if (s[i] == '\r' && s[i + 1] == '\n')
          {
            s += 2;
            //*(result++) = '\r';
            //*(result++) = '\n';
            break;
          }
          else
          {
            //we have an error in the encoding...
            status = 1;
            //s--;
          }
        }
        mid[i] = s[i];
      }
      //now we just have to convert the hex string to an char...
      if (ok)
      {
        s += 2;
        int m = hexmap[mid[0]];
        m <<= 4;
        m |= hexmap[mid[1]];
        *(result++) = (char)m;
      }
    }
    else
    {
      if (*s != '\0') *(result++) = *(s++);
    }
  }
  strcpy(line,finalresult);
  delete finalresult;
  return strlen(line);
}
//---------------------------------------------------------------------------
bool __fastcall TMIMEdecoder::DecodeBase64Line(AnsiString &line, TStream *stream)
{
   if (mimedlg && mimedlg->Caption.Pos("base64")==0) mimedlg->Caption = "Decoding base64...";
   strcpy(buffer,line.c_str());
   int size=DecodeBase64Line(buffer);
   stream->WriteBuffer(buffer,size);
   return status==statMimeOk;
}
//---------------------------------------------------------------------------
int __fastcall TMIMEdecoder::DecodeBase64Line(char *line)
{
   ok = true; status = statMimeOk;
   byte ch,inbuf[3],outbuf[4];
   bool flignore,flendtext=false;
   unsigned ixtext=0,oxtext=0,lentext=strlen(line);
   short ixinbuf=0;

   while(ixtext<lentext)
   {
      if (cancel)
      {
         status = statMimeCancel;
         ok = false;
         return -1;
      }
      ch = line[ixtext++];
      flignore = false;

      if ((ch >= 'A') && (ch <= 'Z')) ch = (byte)(ch - 'A'); else
      if ((ch >= 'a') && (ch <= 'z')) ch = (byte)(ch - 'a' + 26); else
      if ((ch >= '0') && (ch <= '9')) ch = (byte)(ch - '0' + 52); else
      if ( ch == '+') ch = 62; else
      if ( ch == '=') flendtext = true; else
      if ( ch == '/') ch = 63; else
         flignore = true;

      if (!flignore)
      {
         short charsinbuf = 3;
         boolean flbreak = false;
         if (flendtext)
         {
            if (ixinbuf==0) break;

            if ((ixinbuf==1) || (ixinbuf==2)) charsinbuf = 1;
            else charsinbuf = 2;

            flbreak = true;
            ixinbuf = 3;
         }

         inbuf[ixinbuf++] = ch;
         if (ixinbuf==4)
         {
            ixinbuf = 0;
            outbuf[0] = (byte)((inbuf[0] << 2) | ((inbuf[1] & 0x30) >> 4));
            outbuf[1] = (byte)(((inbuf[1] & 0x0F) << 4) | ((inbuf[2] & 0x3C) >> 2));
            outbuf[2] = (byte)(((inbuf[2] & 0x03) << 6) | ( inbuf[3] & 0x3F));
            for (int i=0; i<charsinbuf; i++)
               line[oxtext++] = outbuf[i];
         }
         if (flbreak) break;
      }
   }
   line[oxtext] = '\0';
   return oxtext;
}
//---------------------------------------------------------------------------
bool __fastcall TMIMEdecoder::DecodeUUencodeLine(AnsiString &line, TStream *stream)
{
   if (mimedlg && mimedlg->Caption.Pos("uuencode")==0) mimedlg->Caption = "Decoding x-uuencode...";
   strcpy(buffer,line.c_str());
   int size=DecodeUUencodeLine(buffer);
   if (!cancel && size>0) stream->WriteBuffer(buffer,size);
   return status==statMimeOk;
}
//---------------------------------------------------------------------------
#define DEC(c)	(((c) - ' ') & 077) // single character decode
int  __fastcall TMIMEdecoder::DecodeUUencodeLine(char *line)
{
    ok = true; status = statMimeOk;
    if (s(line).Pos("begin ")==1 || s(line).Pos("end")==1) return 0;
    int n,c1,c2,c3;
    const int bufsize=80;
	char buf[bufsize],*bp,ch;
    unsigned ixtext=0,oxtext=0,lentext=strlen(line);

	while(ixtext<lentext)
    {
        if (cancel) return -1;
        for (int i=0; i<bufsize; i++)
        {
           if (ixtext>=lentext) break;
           ch = line[ixtext++];
           if (ch=='\r') continue;
           if (ch=='\n') break;
           buf[i] = ch;
        }

		n = DEC(buf[0]); if (n<=0) break;
		bp = &buf[1];
		while (n > 0)
        {
           c1 = ((DEC(bp[0]) << 2) | (DEC(bp[1]) >> 4));
           c2 = ((DEC(bp[1]) << 4) | (DEC(bp[2]) >> 2));
           c3 = ((DEC(bp[2]) << 6) |  DEC(bp[3]));
           if (n>=1) line[oxtext++] = (char)c1;
           if (n>=2) line[oxtext++] = (char)c2;
           if (n>=3) line[oxtext++] = (char)c3;
		   bp += 4;
		   n -= 3;
		}
	}
    line[oxtext] = '\0';
    return oxtext;
}
//---------------------------------------------------------------------------
bool __fastcall TMIMEdecoder::DecodeMessage()
{
   return DecodeFirstPart();
}
//---------------------------------------------------------------------------
AnsiString __fastcall AddSemiColon(AnsiString &line)
{
   if (line.Length() && line[line.Length()]!=';') line += ";";
   return line;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TMIMEdecoder::GetMIMEContent()
{
   bool gotcont=false;
   AnsiString line,content;
   int i=-1; while(++i<mime->Count)
   {
      while(i<mime->Count && (line=mime->Strings[i].LowerCase()).Pos("content-")==0)
         i++;

      if (line.IsEmpty()) break;

      if (line.Pos("content-")==1)
      {
         mime->Delete(i);
         gotcont = true;
         if (line.Pos("content-")>0) content = AddSemiColon(content)+" "+line;
         while(i<mime->Count)
         {
            line = mime->Strings[i].LowerCase();
            if (line.Trim().IsEmpty()) break;

            if (s(" \t").Pos(line[1])>0 || line.Pos("content-")==1)
            {
               AddSemiColon(content);
               content += " "+line.Trim();
               mime->Delete(i);
               gotcont = true;
               continue;
            }
            gotcont = true;
            break;
         }
      }
      if (gotcont)
         break;
   }
   content = content.Trim();
   return content;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TMIMEdecoder::GetPrimeMIMEContent()
{
   AnsiString line,content;
   int i=-1; while(++i<mime->Count)
   {
      if ((line=mime->Strings[i].Trim().LowerCase()).IsEmpty()) break;
      if (line.Pos("content-")==0 && line.Pos("boundary=")==0)
         continue;

      content = AddSemiColon(content)+" "+line;
      while(++i<mime->Count)
      {
         if ((line=mime->Strings[i].Trim().LowerCase()).IsEmpty()) break;
         if (s(" \t").Pos(line[1])>0)
         {
            content = AddSemiColon(content)+" "+line.Trim();
            continue;
         }
         i--;
         break;
      }
      if (line.IsEmpty())
         break;
   }
   content = content.Trim();
   return content;
}
/*
AnsiString __fastcall TMIMEdecoder::GetPrimeMIMEContent()
{
   AnsiString line,content;
   int i=-1; while(++i<mime->Count)
   {
      if ((line=mime->Strings[i].Trim().LowerCase()).IsEmpty()) break;
      if (line.Pos("content-")==0 && line.Pos("boundary=")==0)
         continue;

      content = AddSemiColon(content)+" "+line;
      while(++i<mime->Count)
      {
         if ((line=mime->Strings[i].Trim().LowerCase()).IsEmpty()) break;
         if (s(" \t").Pos(line[1])>0 || line.Pos("content-")==1 || line.Pos("boundary=")==1)
         {
            content = AddSemiColon(content)+" "+line.Trim();
            continue;
         }
         break;
      }
   }
   content = content.Trim();
   return content;
}*/
//---------------------------------------------------------------------------
AnsiString __fastcall TMIMEdecoder::GetMIMEVersion()
{
   AnsiString line,version;
   for (int i=0; i<mime->Count; i++)
   {
      line = mime->Strings[i].LowerCase();
      if (line.Trim().IsEmpty()) break;
      if (line.Pos("mime-version: ")==1)
      {
         version = line.SubString(14,16);
         //mime->Delete(i);
         break;
      }
   }
   return version;
}
//---------------------------------------------------------------------------
bool __fastcall TMIMEdecoder::DecodeUncodedLine(AnsiString &line, TStream *stream)
{
   if (mimedlg && mimedlg->Caption!="Decoding...") mimedlg->Caption = "Decoding...";
   ok = true; status = statMimeOk;
   if (contenthead->Count)
   {
      AnsiString header = contenthead->Strings[0];
      if (header.IsEmpty() || header.Pos("--")==1) contenthead->Delete(0);
      contenthead->Add("");

      strcpy(buffer,contenthead->Text.c_str());
      stream->WriteBuffer(buffer,strlen(buffer));
      contenthead->Clear();
   }
   stream->WriteBuffer(s(line+"\r\n").c_str(),line.Length()+2);
   return true;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TMIMEdecoder::AttachmentFilename(AnsiString &contentline)
{
   AnsiString line;
   int i=-1; while(!(line=GetToken(++i,contentline,";").Trim()).IsEmpty())
   {
      if (line.Trim().LowerCase().Pos("filename=")==1)
         return AddBS(attachdir)+dropquotes(line.SubString(10,1024).Trim());
   }
   BingBox((mimedlg?mimedlg:NULL),
      s("I have encountered an attachment that has no accompanying\n")+
      s("filename. I'll save this as 'noname.txt'."),
      s("Filename Error"), MB_OK);
   return AddBS(attachdir)+"noname.txt";
}
//---------------------------------------------------------------------------
AnsiString __fastcall TMIMEdecoder::ContentEncoding(AnsiString &contentline)
{
   AnsiString line;
   int i=-1; while(!(line=GetToken(++i,contentline,";").Trim()).IsEmpty())
   {
      if (line.Trim().LowerCase().Pos("content-transfer-encoding:")==1)
         return dropquotes(line.SubString(27,1024).Trim());
   }
   return "";
}
//---------------------------------------------------------------------------
AnsiString __fastcall TMIMEdecoder::ContentBoundary(AnsiString &contentline)
{
   AnsiString line;
   int i=-1; while(!(line=GetToken(++i,contentline,";").Trim()).IsEmpty())
   {
      if (line.Trim().LowerCase().Pos("boundary=")==1)
         return dropquotes(line.SubString(14,1024).Trim());
   }
   return "";
}
//---------------------------------------------------------------------------
AnsiString __fastcall TMIMEdecoder::ContentType(AnsiString &contentline)
{
   AnsiString line;
   int i=-1; while(!(line=GetToken(++i,contentline,";").Trim()).IsEmpty())
   {
      if (line.Trim().LowerCase().Pos("content-type:")==1)
         return dropquotes(line.SubString(14,1024).Trim());
   }
   return "";
}
//---------------------------------------------------------------------------
bool __fastcall TMIMEdecoder::IsAttachment(AnsiString &contentline)
{
   return contentline.Pos("content-disposition: attachment")>0;
}
//---------------------------------------------------------------------------
